from Acquisition import aq_inner
from Products.Archetypes.interfaces import IBaseObject
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import PloneMessageFactory as _
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from plone.app.linkintegrity.handlers import modifiedArchetype


class UpdateView(BrowserView):

    def __call__(self):
        context = aq_inner(self.context)
        request = aq_inner(self.request)
        clicked = request.form.has_key
        if clicked('update') or clicked('delete_all'):
            count = self.update()
            msg = _(u'linkintegrity_update_info',
                default=u'Link integrity information updated for ${count} item(s).',
                mapping={'count': count})
            IStatusMessage(request).addStatusMessage(msg, type='info')
            request.RESPONSE.redirect(getToolByName(context, 'portal_url')())
        elif clicked('cancel'):
            msg = _(u'Update cancelled.')
            IStatusMessage(request).addStatusMessage(msg, type='info')
            request.RESPONSE.redirect(getToolByName(context, 'portal_url')())
        else:
            return self.index()

    def update(self):
        catalog = getToolByName(self.context, 'portal_catalog')
        count = 0
        for brain in catalog(Language='all'):
            obj = brain.getObject()
            if IBaseObject.providedBy(obj):
                modifiedArchetype(obj, 'dummy event parameter')
                count += 1
        return count




from Products.Archetypes.interfaces import IReference
from Products.Archetypes.Field import TextField
from Products.Archetypes.exceptions import ReferenceException
from OFS.interfaces import IItem
from exceptions import LinkIntegrityNotificationException
from interfaces import ILinkIntegrityInfo, IOFSImage
from urlparse import urlsplit, urlunsplit
from parser import extractLinks

referencedRelationship = 'isReferencing'


def findObject(base, path):
    """ traverse to given path and find the upmost object """
    obj = base
    components = path.split('/')
    while components:
        try: child = obj.restrictedTraverse(components[0])
        except: return None, None
        if not IItem.providedBy(child):
            break
        obj = child
        components.pop(0)
    return obj, '/'.join(components)


def getObjectsFromLinks(base, links):
    """ determine actual objects (plus extra paths) refered to by given links """
    objects = []
    url = base.absolute_url()
    scheme, host, path, query, frag = urlsplit(url)
    site = urlunsplit((scheme, host, '', '', ''))
    for link in links:
        s, h, path, q, f = urlsplit(link)
        if (not s and not h) or (s == scheme and h == host):    # relative or local url
            obj, extra = findObject(base, path)
            if obj:
                extra += urlunsplit(('', '', '', q, f))
                objects.append((obj, extra))
    return objects


def modifiedArchetype(obj, event):
    """ an archetype based object was modified """
    try:    # TODO: is this a bug or a needed workaround?
        for ref in obj.getReferences(relationship=referencedRelationship):
            # only remove forward references
            obj.deleteReference(ref, relationship=referencedRelationship)
    except AttributeError:
        return
    refs = []
    for field in obj.Schema().fields():
        if isinstance(field, TextField):
            accessor = field.getAccessor(obj)
            links = extractLinks(accessor())
            refs.extend(getObjectsFromLinks(obj, links))
    for ref, extra in refs:
        if IOFSImage.providedBy(ref):
            ref = ref.aq_parent     # use atimage object for scaled images
        try:
            obj.addReference(ref, relationship=referencedRelationship)
        except ReferenceException:
            pass


def referenceRemoved(obj, event):
    """ store information about the removed link integrity reference """
    assert IReference.providedBy(obj)
    assert obj is event.object          # just making sure...
    if not obj.relationship == referencedRelationship:
        return                          # skip for other removed references
    # if the object the event was fired on doesn't have a `REQUEST` attribute
    # we can safely assume no direct user action was involved and therefore
    # never raise a link integrity exception...
    if not hasattr(obj, 'REQUEST'):
        return
    storage = ILinkIntegrityInfo(obj.REQUEST)
    breaches = storage.getIntegrityBreaches()
    breaches.setdefault(obj.getTargetObject(), set()).add(obj.getSourceObject())
    storage.setIntegrityBreaches(breaches)


def referencedObjectRemoved(obj, event):
    """ check if the removal was already confirmed or redirect to the form """
    # if the object the event was fired on doesn't have a `REQUEST` attribute
    # we can safely assume no direct user action was involved and therefore
    # never raise a link integrity exception...
    # (this should also fix http://plone.org/products/cachefu/issues/86)
    if not hasattr(obj, 'REQUEST'):
        return
    info = ILinkIntegrityInfo(obj.REQUEST)

    # first we check if link integrity checking was enabled
    if not info.integrityCheckingEnabled():
        return

    # since the event gets called for every subobject before it's
    # called for the item deleted directly via _delObject (event.object)
    # itself, but we do not want to present the user with a confirmation
    # form for every (referred) subobject, so we remember and skip them...
    info.addDeletedItem(obj)
    if obj is not event.object:
        return

    # if the number of expected events has been stored to help us prevent
    # multiple forms (i.e. in folder_delete), we wait for the next event
    # if we know there will be another...
    if info.moreEventsToExpect():
        return

    # at this point all subobjects have been removed already, so all
    # link integrity breaches caused by that have been collected as well;
    # if there aren't any (after things have been cleaned up),
    # we keep lurking in the shadows...
    if not info.getIntegrityBreaches():
        return

    # if the user has confirmed to remove the currently handled item in a
    # previous confirmation form we won't need it anymore this time around...    
    if info.isConfirmedItem(obj):
        return

    # otherwise we raise an exception and pass the object that is supposed
    # to be removed as the exception value so we can use it as the context
    # for the view triggered by the exception;  this is needed since the
    # view is an adapter for the exception and a request, so it gets the
    # exception object as the context, which is not very useful...
    raise LinkIntegrityNotificationException, obj




from Acquisition import aq_inner
from Products.Five import BrowserView
from Products.CMFCore.utils import getToolByName, _checkPermission
from Products.CMFCore.permissions import AccessContentsInformation

from plone.app.linkintegrity.interfaces import ILinkIntegrityInfo
from plone.app.linkintegrity.utils import encodeRequestData


class RemoveConfirmationView(BrowserView):

    def __init__(self, context, request):
        # since this is a view adapting an exception and a request (instead
        # of a regular content object and a request), the exception object
        # was passed as the context;  therefore we need to construct a
        # proper context in order to render the template in a sane manner;
        # to do this we will assume that the desired context was passed as
        # the first exception value...
        self.exception = context
        self.context, = context.args
        self.request = request

    def originalRequest(self):
        # in order to interrupt the current request with a confirmation
        # question about removing the referred to object we need to save
        # the original request to be able to possibly continue it later on,
        # so we pickle and encode its body and environment...
        self.request.stdin.seek(0)
        body = self.request.stdin.getvalue()    # zope2 request body...
        env = dict(self.request._orig_env)
        for key in 'HTTP_AUTHORIZATION', 'HTTP_COOKIE':
            if env.has_key(key):
                del env[key]
        return encodeRequestData((body, env))

    def integrityBreaches(self):
        info = ILinkIntegrityInfo(self.request).getIntegrityBreaches()
        byTitle = lambda a,b: cmp(a.Title(), b.Title())
        breaches = []
        for target, sources in info.items():
            breaches.append({
                'title': target.Title(),
                'type': target.getPortalTypeName(),
                'url': target.absolute_url(),
                'sources': sorted(sources, byTitle),
            })
        return sorted(breaches, lambda a,b: cmp(a['title'], b['title']))

    def isAccessible(self, obj):
        return _checkPermission(AccessContentsInformation, obj)

    def confirmedItems(self):
        info = ILinkIntegrityInfo(self.request)
        targets = info.getIntegrityBreaches().keys()
        return info.encodeConfirmedItems(additions=targets)

    def callbackURL(self):
        portal = getToolByName(aq_inner(self.context), "portal_url")
        return portal() + '/removeConfirmationAction'

    def cancelURL(self):
        url = self.request.environ.get('HTTP_REFERER', None)
        if url is None:
            url = getToolByName(aq_inner(self.context), "portal_url")()
        return url




from zope.interface import implements
from interfaces import ILinkIntegrityNotificationException
from OFS.ObjectManager import BeforeDeleteException


class LinkIntegrityNotificationException(BeforeDeleteException):
    """ an exception indicating a prevented link integrity breach """
    implements(ILinkIntegrityNotificationException)





from base64 import b64encode, b64decode
from zlib import compress, decompress


def encodeStrings(strings):
    """ compress and encode a list of strings into a single string """
    def _encode(strings):
        for string in strings:
            yield '%d:%s' % (len(string), string)
    return b64encode(compress(''.join(_encode(strings))))


def decodeStrings(data):
    """ decode and uncompress a string as a generator """
    data = decompress(b64decode(data))
    while data:
        pos = data.find(':')
        end = pos + int(data[:pos]) + 1
        yield data[pos+1:end]
        data = data[end:]


def encodeRequestData((body, env)):
    """ encode the relevant request data, i.e. body and env """
    def _iterdata():
        yield body
        for key, val in env.iteritems():
            yield str(key)
            yield str(val)
    return encodeStrings(_iterdata())


def decodeRequestData(data):
    """ decode the relevant request data, i.e. body and env """
    def _pertwo():
        while True:
            yield data.next(), data.next()
    data = decodeStrings(data)
    return data.next(), dict(_pertwo())


def encodeInts(ints):
    """ encode a list of integers into a string """
    return ':'.join(map(str, ints))


def decodeInts(data):
    """ decode a string to a list of integers """
    return map(int, data.split(':'))




from zope.interface import Interface


class ILinkIntegrityNotificationException(Interface):
    """ an exception indicating a prevented link integrity breach """


class ILinkIntegrityTool(Interface):
    """ interface of link integrity tool """


class ILinkIntegrityInfo(Interface):
    """ a place to store information about link integrity, i.e. breaches;
        the storage is assumed to be unique per browser request """

    def getIntegrityBreaches():
        """ return stored information regarding link integrity breaches """

    def setIntegrityBreaches(breaches):
        """ store information regarding link integrity breaches """

    def getDeletedItems():
        """ return information about all items deleted during the request """

    def addDeletedItem(item):
        """ remember an item deleted during the request """

    def getEnvMarker():
        """ return the marker string used to pass the already confirmed
            items across the retry exception """

    def isConfirmedItem(obj):
        """ indicate if the removal of the given object was confirmed """

    def encodeConfirmedItems(additions):
        """ return the list of previously confirmed (for removeal) items,
            optionally adding the given items, encoded for usage in a form """


class IOFSImage(Interface):
    """ interface for OFS.Image.Image """




# the following code is "stolen" from (or maybe it was inspired by? :))
# FiveException (http://codespeak.net/svn/z3/FiveException)

from zope.component import queryMultiAdapter
from Zope2.App.startup import zpublisher_exception_hook
from ZPublisher.Publish import Retry


def zpublisher_exception_hook_wrapper(published, REQUEST, t, v, traceback):
    """ wrapper around the zope2 zpublisher's error hook """
    try:
        # if we got a retry exception, we just propagate it instead of
        # trying to log it (like FiveException does)
        if t is Retry:
            v.reraise()
        # first we try to find a view/adapter for the current exception and
        # let the original function try to handle the exception if we can't
        # find one...
        view = queryMultiAdapter((v, REQUEST), name='index.html', default=None)
        if view is None:
            zpublisher_exception_hook(published, REQUEST, t, v, traceback)
        else:
            # otherwise render the view and raise the rendered string like
            # raise_standardErrorMessage does...
            view = view.__of__(published)
            message = view()
            if isinstance(message, unicode):
                message = message.encode('utf-8')
            raise t, message, traceback
    finally:
        traceback = None


from ZPublisher.Publish import get_module_info
def proxy_get_module_info(*args, **kwargs):
    results = list(get_module_info(*args, **kwargs))
    if results[5] is zpublisher_exception_hook:
        results[5] = zpublisher_exception_hook_wrapper
    return tuple(results)


def installExceptionHook():
    import ZPublisher.Publish
    ZPublisher.Publish.get_module_info = proxy_get_module_info


def retry(self):
    """ re-initialize a response object to be used in a retry attempt """
    # this implementation changes the original one so that the response
    # instance is reused instead of replaced with a new one (after a Retry
    # exception was raised);  this fixes a bug in zopedoctests' http()
    # function (Testing/ZopeTestCase/zopedoctest/functional.py:113);
    # the doctest code assumes that the HTTPResponse instance passed to
    # publish_module() (line 177) is used to handle to complete request, so
    # it can be used to get the status, headers etc later on (lines 183-186);
    # normally this is okay, but raising a Retry will create a new response
    # instance, which will then hold that data (relevant for evaluating the
    # doctest) while the original (passed in) instance is still empty...
    #
    # so to fix this (quickly) retry() now cleans up and returns itself:
    self.__init__(stdout=self.stdout, stderr=self.stderr)
    return self

from ZPublisher.HTTPResponse import HTTPResponse
HTTPResponse.retry = retry




from AccessControl.SecurityManagement import newSecurityManager
from Testing import ZopeTestCase
from Products.PloneTestCase import PloneTestCase
from Products.PloneTestCase.layer import PloneSite
from StringIO import StringIO
from base64 import decodestring
from transaction import commit


class PloneLinkintegrity(PloneSite):

    @classmethod
    def setUp(cls):
        app = ZopeTestCase.app()
        portal = app.plone

        # login as admin (copied from `loginAsPortalOwner`)
        uf = app.acl_users
        user = uf.getUserById(PloneTestCase.portal_owner).__of__(uf)
        newSecurityManager(None, user)

        # create sample content
        gif = 'R0lGODlhAQABAPAAAPj8+AAAACH5BAEAAAAALAAAAAABAAEAAAICRAEAOw=='
        gif = StringIO(decodestring(gif))
        portal.invokeFactory('Document', id='doc1', title='Test Page 1',
            text='<html> <body> a test page </body> </html>')
        portal.invokeFactory('Document', id='doc2', title='Test Page 2',
            text='<html> <body> another test page </body> </html>')
        portal.invokeFactory('Image', id='image1', title='Test Image 1', image=gif)
        portal.invokeFactory('Image', id='image2', title='Test Image 2', image=gif)
        portal.invokeFactory('Image', id='image3', title='Test Image 3', image=gif)
        portal.invokeFactory('Folder', id='folder1', title='Test Folder 1')
        portal.folder1.invokeFactory('Document', id='doc3', title='Test Page 3',
            text='<html> <body> a test page in a subfolder </body> </html>')
        portal.folder1.invokeFactory('Document', id='doc4', title='Test Page 4',
            text='<html> <body> another test page </body> </html>')
        portal.folder1.invokeFactory('Document', id='doc5', title='Test Page 5',
            text='<html> <body> another test page </body> </html>')

        # starting with 2.10.4 product initialization gets delayed for
        # instance startup and is never called when running tests;  hence
        # we have to initialize the package method manually...
        from OFS.Application import install_package
        import plone.app.linkintegrity
        install_package(app, plone.app.linkintegrity, plone.app.linkintegrity.initialize)

        # create a starting point for the tests...
        commit()
        ZopeTestCase.close(app)

    @classmethod
    def tearDown(cls):
        app = ZopeTestCase.app()
        portal = app.plone

        # login as admin (copied from `loginAsPortalOwner`)
        uf = app.acl_users
        user = uf.getUserById(PloneTestCase.portal_owner).__of__(uf)
        newSecurityManager(None, user)

        # remove sample content
        ids = 'doc1', 'doc2', 'image1', 'image2', 'image3', 'folder1'
        portal.manage_delObjects(ids=list(ids))

        # commit the cleanup...
        commit()
        ZopeTestCase.close(app)


